from os.path import join
from unittest import TestCase

from ewconfig.merge.prefer import Prefer
from ewconfig.merge.slink import SlinkParser
from ..lib.file import project_root
from ..lib.log import LogMixin


class TestSlink(LogMixin, TestCase):

    stderr_verbosity = 5

    def test_parse(self):
        slink = SlinkParser('slink_imports.d', Prefer.CHECK, True)
        slink.read(join(project_root(), 'data', 'reference'))
        self.assertEqual(slink._selectors, ['BH?.D'])
        self.assertEqual(slink._streams, {'GE_APE': [],
                                          'GE_DSB': ['BH?.D', 'LH?.D'],
                                          'GE_ISP': [],
                                          'GE_STU': [],
                                          'TA_*': []})

    def test_idempotent(self):
        old = SlinkParser('slink_imports.d', Prefer.CHECK, True)
        old.read(join(project_root(), 'data', 'reference'))
        new = SlinkParser('slink_imports.d', Prefer.CHECK, True)
        new.read(join(project_root(), 'data', 'reference'))
        old.add(new)
        self.assertEqual(new._selectors, old._selectors)
        self.assertEqual(new._streams, old._streams)

    def test_merge_conflict(self):
        old = SlinkParser('slink_imports.d', Prefer.CHECK, True)
        old.read(join(project_root(), 'data', 'reference'))
        new = SlinkParser('slink_imports-conflict.d', Prefer.CHECK, True)
        new.read(join(project_root(), 'data', 'reference'))
        with self.assertRaisesRegex(Exception, 'Inconsistent selectors'):
            old.add(new)

    def test_merge_old(self):
        old = SlinkParser('slink_imports.d', Prefer.OLD, True)
        old.read(join(project_root(), 'data', 'reference'))
        new = SlinkParser('slink_imports-conflict.d', Prefer.OLD, True)
        new.read(join(project_root(), 'data', 'reference'))
        old.add(new)
        old_old = SlinkParser('slink_imports.d', Prefer.OLD, True)
        old_old.read(join(project_root(), 'data', 'reference'))
        self.assertEqual(old_old._selectors, old._selectors)
        self.assertEqual(old_old._streams, old._streams)

    def test_merge_new(self):
        old = SlinkParser('slink_imports.d', Prefer.NEW, True)
        old.read(join(project_root(), 'data', 'reference'))
        new = SlinkParser('slink_imports-conflict.d', Prefer.NEW, True)
        new.read(join(project_root(), 'data', 'reference'))
        old.add(new)
        old_new = SlinkParser('slink_imports-conflict.d', Prefer.NEW, True)
        old_new.read(join(project_root(), 'data', 'reference'))
        self.assertEqual(old_new._selectors, old._selectors)
        self.assertEqual(old_new._streams, old._streams)




from io import StringIO
from logging import StreamHandler, getLogger

from sys import stderr

from ewconfig.lib.log import make_log, clear_log


class LogMixin:

    stderr_verbosity = 4

    def setUp(self):
        stderr_verbosity = self.stderr_verbosity
        print(file=stderr)  # sabotage unittest's assumptions about inline stderr printing
        make_log(self.__class__.__name__, stderr_verbosity, path='/tmp')
        # do super (and sibling) classes AFTER setting up logs
        super().setUp()

    def tearDown(self):
        super().tearDown()
        clear_log()


class CaptureLog:
    '''
    https://stackoverflow.com/a/63678561
    '''

    def __init__(self, name='ewconfig'):
        self.logger = getLogger(name)
        self.io = StringIO()
        self.sh = StreamHandler(self.io)
        self.out = ''

    def __enter__(self):
        self.logger.addHandler(self.sh)
        return self

    def __exit__(self, *exc):
        self.logger.removeHandler(self.sh)
        self.out = self.io.getvalue()

    def __repr__(self):
        return f'captured: "{self.out}"\n'



from fnmatch import translate
from logging import getLogger
from re import compile

log = getLogger(__name__)


class Filter:

    def __init__(self, name, includes=None, excludes=None):
        '''
        Create a filter for <name> that includes patterns in <includes>
        and then excludes patterns in <excludes>.
        If <includes> is missing then all are initially included.
        If <excludes> is missing then none are excluded.
        Patterns are 'globs' as supported by fnmatch module.
        '''
        self.__name = name
        if isinstance(includes, str) or isinstance(excludes, str):
            raise Exception('Filter takes lists of includes and excludes')
        if not includes: includes = ['*']
        if not excludes: excludes = []
        self.__str = f'{name}: include ' + ','.join(includes)
        if excludes: self.__str += '; exclude ' + ','.join(includes)
        self.__includes = [compile(translate(include)) for include in includes]
        self.__excludes = [compile(translate(exclude)) for exclude in excludes]

    def __call__(self, items, key=None, quiet=False):
        logged = set()
        if not key: key = lambda x: x
        for item in items:
            value = key(item)
            if any(include.match(value) for include in self.__includes):
                if not any(exclude.match(value) for exclude in self.__excludes):
                    yield item
                else:
                    if value not in logged:
                        if not quiet: log.info(f'{self.__name} {value} was explicitly excluded')
                        logged.add(value)
            else:
                if value not in logged:
                    if not quiet: log.info(f'{self.__name} {value} was not included')
                    logged.add(value)

    def __str__(self):
        return self.__str




from filecmp import cmp
from logging import getLogger
from os import listdir, makedirs
from os.path import abspath, normpath, expanduser, exists, isdir
from tempfile import mkdtemp

log = getLogger(__name__)


def assert_dir(dir, name=None, create=False):
    dir = clean_path(dir)
    if not exists(dir):
        if create:
            log.debug(f'Creating {dir}')
            makedirs(dir)
        else:
            raise Exception(f'{dir} {"(" + name + ") " if name else ""}does not exist')
    if not isdir(dir):
        raise Exception(f'{dir} ({name}) is not a directory')
    log.debug(f'{dir} exists and is a directory')
    return dir


def tmp_dir():
    # note this does not delete or clean up the directory, it just creates it
    return mkdtemp(prefix='ewconfig')


def assert_empty_dir(dir, name=None):
    dir = assert_dir(dir, name=name, create=True)
    if listdir(dir):
        raise Exception(f'{dir} {"(" + name + ") " if name else ""}is not empty')
    log.debug(f'{dir} is empty')
    return dir


def clean_path(path):
    return abspath(normpath(expanduser(path)))


def assert_equal_contents(file1, file2):
    if not cmp(file1, file2, shallow=False):
        raise Exception(f'Files {file1} and {file2} differ')
    else:
        log.debug(f'Compared {file1} and {file2}')



from itertools import zip_longest
from logging import getLogger
from os import makedirs, chdir, listdir
from os.path import join, exists
from re import search, sub, compile
from shutil import rmtree
from subprocess import run

from ewconfig.lib.file import tmp_dir
from .compare import Compare
from .file import project_root

log = getLogger(__name__)


class MainMixin(Compare):

    def _src_dir(self, name):
        return join(project_root(), 'data', name.lower())

    def _test_dir(self, name):
        return join(self._src_dir(name), 'x2ew')

    def _dest_dir(self, name):
        return join(tmp_dir(), name.lower())

    def _create_dest_dir(self, name):
        dest_dir = self._dest_dir(name)
        rmtree(dest_dir, ignore_errors=True)
        makedirs(dest_dir)
        chdir(dest_dir)
        return dest_dir

    def run_sac2ew(self, name):
        from ewconfig.sacpz2ew import main
        dest_dir = self._create_dest_dir(name)
        src_dir = self._src_dir(name)
        log.info(f'Running with input from {src_dir} and output to {dest_dir}')
        log.debug('Calling rdseed')
        run(['rdseed', '-pf', join(src_dir, name + '.dataless')])
        main(dest_dir)
        return dest_dir

    def run_xml2ew(self, name, **kwargs):
        from ewconfig.stationxml2ew import main_single
        dest_dir = self._create_dest_dir(name)
        src_dir = self._src_dir(name)
        log.info(f'Running with input from {src_dir} and output to {dest_dir}')
        main_single(join(src_dir, name + '.xml'), **kwargs)
        return dest_dir

    def assert_files(self, dest_dir, name, rdseed=False, skip_files=None, skip_lines=None):
        test_dir = self._test_dir(name)
        self._assert_files(dest_dir, test_dir, rdseed=rdseed, skip_files=skip_files, skip_lines=skip_lines)



from shutil import rmtree
from unittest import TestCase

from ewconfig.lib.filter import Filter
from ..lib.x2ew import MainMixin
from ..lib.log import LogMixin


class TestR50A(MainMixin, LogMixin, TestCase):

    stderr_verbosity = 5

    def test_original(self):
        dest_dir = self.run_sac2ew('R50A')
        # fails because we have no way of restricting date range
        with self.assertRaises(AssertionError):
            self.assert_files(dest_dir, 'R50A', rdseed=True)
        rmtree(dest_dir)

    def test_station_xml(self):
        # note that we need to supply default date (not supplied by ArgsParser)
        dest_dir = self.run_xml2ew('R50A', chan_filter=Filter('Channel', excludes=['LOG']), lo_precision=True)
        self.assert_files(dest_dir, 'R50A')
        rmtree(dest_dir)



from enum import Enum

from .pz import PolesZeros


class State(Enum):
    NORMAL = 0
    POLES = 1
    ZEROS = 2


class SacPolesZeros(PolesZeros):

    def __init__(self, file):
        super().__init__()
        self.__num_poles = 0
        self.__num_zeros = 0
        self.__parse(file)

    def __parse(self, file):
        with open(file, 'r') as lines:
            state = State.NORMAL
            for line in lines:
                if state == State.POLES:
                    self.poles_lines.append(line)
                    if len(self.poles_lines) == self.__num_poles:
                        state = State.NORMAL
                    continue
                if state == State.ZEROS:
                    # in zero gathering state
                    self.zeros_lines.append(line)
                    if len(self.zeros_lines) == self.__num_zeros:
                        state = State.NORMAL
                    continue
                a = line.strip().split()
                if a[0] == 'CONSTANT':
                    self.constant = float(a[1])
                    continue
                if a[0] == 'POLES':
                    self.__num_poles = int(a[1])
                    state = State.POLES if self.__num_poles else State.NORMAL
                    continue
                if a[0] == 'ZEROS':
                    self.__num_zeros = int(a[1])
                    state = State.ZEROS if self.__num_zeros else State.NORMAL
                    continue
                if a[0] == '*':
                    # a comment, check for lat, lon, elev, SCNL
                    self.keep_comment = self.keep_comment + line;
                    if a[1] == 'NETWORK':
                        self.net = a[3]
                        continue
                    if a[1] == 'STATION':
                        self.sta = a[3]
                        continue
                    if a[1] == 'CHANNEL':
                        self.chan = a[3]
                        continue
                    if a[1] == 'LOCATION':
                        if len(a) == 4:
                            self.loc = a[3]
                        else:
                            self.loc = '--'
                        continue
                    if a[1] == 'LATITUDE':
                        self.lat = float(a[3])
                        continue
                    if a[1] == 'LONGITUDE':
                        self.lon = float(a[3])
                        continue
                    if a[1] == 'ELEVATION':
                        self.elev = float(a[3])
                        continue




from logging import getLogger
from shutil import rmtree
from unittest import TestCase

from ewconfig.merge.prefer import Prefer
from ..lib.log import LogMixin
from ..lib.merge import MainMixin

log = getLogger(__name__)


class TestIssue6(MainMixin, LogMixin, TestCase):

    stderr_verbosity = 5

    def test_merge(self):
        merged = self.run_merge('issue-6-a', ['issue-6-b'], sort=False)
        self.assert_files(merged, 'issue-6-a')
        rmtree(merged)

    def test_conflict_error(self):
        with self.assertRaisesRegex(Exception, 'Entries differ'):
            merged = self.run_merge('issue-6-a', ['issue-6-c'])

    def test_conflict_a(self):
        merged = self.run_merge('issue-6-a', ['issue-6-c'], prefer=Prefer.OLD, sort=False)
        self.assert_files(merged, 'issue-6-a')
        rmtree(merged)

    def test_conflict_abc(self):
        merged = self.run_merge('issue-6-a', ['issue-6-b', 'issue-6-c'], prefer=Prefer.OLD, sort=False)
        self.assert_files(merged, 'issue-6-a')
        rmtree(merged)

    def test_conflict_c(self):
        merged = self.run_merge('issue-6-a', ['issue-6-c'], prefer=Prefer.NEW, sort=False)
        self.assert_files(merged, 'issue-6-c')
        rmtree(merged)



from logging import getLogger

from .prefer import Prefer

log = getLogger(__name__)


def merge(old, new, merged, prefer, parser, assert_eq, writer):
    '''
    This is a HOF that merges old and new data in to merged.  It assumes that parser returns
    a map from NSCL to some info that is used by assert_eq and writer.  Unfortunately it's only
    used by SAC files because it turns out that other files have extra info that we need to
    carry across.
    '''
    nscls_map = merge_nscls(old, new, prefer, parser, assert_eq)
    writer(merged, nscls_map)


def merge_nscls(old, new, prefer, parser, assert_eq):
    old_nscls_map = parser(old)
    new_nscls_map = parser(new)
    if prefer == Prefer.CHECK:
        nscls_map = check_duplicates(old_nscls_map, new_nscls_map, assert_eq)
    else:
        nscls_map = pick_duplicates(old_nscls_map, new_nscls_map, prefer)
    return nscls_map


def check_duplicates(old_nscls_map, new_nscls_map, assert_eq):
    for old_nscl, old_value in old_nscls_map.items():
        if old_nscl in new_nscls_map:
            assert_eq(old_value, new_nscls_map[old_nscl])
            del new_nscls_map[old_nscl]
    old_nscls_map.update(new_nscls_map)
    return old_nscls_map


def pick_duplicates(old_nscls_map, new_nscls_map, prefer):
    conflicts = set(old_nscls_map.keys()).intersection(new_nscls_map.keys())
    if conflicts:
        conflicts_str = ', '.join(conflicts)
        if prefer == Prefer.OLD:
            log.info(f'Will take {conflicts_str} from old config')
            new_nscls_map.update(old_nscls_map)
            return new_nscls_map
        else:
            log.info(f'Will take {conflicts_str} from new config')
            old_nscls_map.update(new_nscls_map)
            return old_nscls_map
    else:
        old_nscls_map.update(new_nscls_map)
        return old_nscls_map



from shutil import rmtree
from unittest import TestCase

from ..lib.x2ew import MainMixin
from ..lib.log import LogMixin


class TestInceV1(MainMixin, LogMixin, TestCase):

    stderr_verbosity = 5

    def test_original(self):
        dest_dir = self.run_sac2ew('Ince_v1')
        self.assert_files(dest_dir, 'Ince_v1', rdseed=True)
        rmtree(dest_dir)

    def test_station_xml(self):
        dest_dir = self.run_xml2ew('Ince_v1')
        self.assert_files(dest_dir, 'Ince_v1')
        rmtree(dest_dir)



from unittest import TestCase

from ewconfig.lib.filter import Filter
from ..lib.log import LogMixin


class TestFilter(LogMixin, TestCase):

    stderr_verbosity = 5

    def test_literal(self):
        self.assertFilter(['a', 'b'], ['b', 'c'], ['a', 'b', 'c'], ['a'])

    def test_default(self):
        self.assertFilter([], ['b', 'c'], ['a', 'b', 'c'], ['a'])

    def test_pattern(self):
        self.assertFilter(['?x*'], ['*z'], ['a', 'x', 'xx', 'xxxx', 'xxxz'], ['xx', 'xxxx'])

    def assertFilter(self, include, exclude, all, target):
        filter = Filter('test', include, exclude)
        result = list(filter(all))
        self.assertEqual(target, result)



'''
a scanner program in python 3 to take SAC PZ files from rdseed -pf and convert
them to localmag and hypoinverse station lists. This is presumed that the 
dataless was complete and had lat/long/elevation for the SCNL

Example output from rdseed -pf
* **********************************
* NETWORK   (KNETWK): TA
* STATION    (KSTNM): R53A
* LOCATION   (KHOLE): 
* CHANNEL   (KCMPNM): BHZ
* CREATED           : 2013-02-05T01:21:15
* START             : 2013-01-13T00:00:00
* END               : 1919-06-29T15:38:39
* DESCRIPTION       : Hurricane, WV, USA
* LATITUDE          : 38.330700 
* LONGITUDE         : -81.951300 
* ELEVATION         : 294.0  
* DEPTH             : 0.0  
* DIP               : 0.0  
* AZIMUTH           : 0.0  
* SAMPLE RATE       : 40.0
* INPUT UNIT        : M
* OUTPUT UNIT       : COUNTS
* INSTTYPE          : Guralp CMG3T/Quanterra 330 Linear Phase Composite
* INSTGAIN          : 1.504200e+03 (M/S)
* COMMENT           : T3P01 0100000A1B8D11B0
* SENSITIVITY       : 6.309070e+08 (M/S)
* A0                : 5.714000e+08
* **********************************
ZEROS   3
        +0.000000e+00   +0.000000e+00
        +0.000000e+00   +0.000000e+00
        +0.000000e+00   +0.000000e+00
POLES   5
        -3.701000e-02   +3.701000e-02
        -3.701000e-02   -3.701000e-02
        -1.131000e+03   +0.000000e+00
        -1.005000e+03   +0.000000e+00
        -5.027000e+02   +0.000000e+00
CONSTANT        +3.605003e+17
'''

from argparse import ArgumentParser
from logging import getLogger
from os import listdir
from os.path import join
from re import search

from .lib.args import add_version_args
from .lib.file import assert_dir
from .lib.log import add_log_args, make_log_from_args
from .lib.markdown import add_md_help_argument
from .lib.write import log_pz, write_all, add_write_args
from .pz.sac import SacPolesZeros

log = getLogger(__name__)


def read(dir, pz):
    files = listdir(dir)
    sacpz_list = []
    for file in files:
        log.info("working on %s", file)
        if search('^SAC_PZs_.*', file):
            sac = pz(join(dir, file))
            log_pz(sac)
            sacpz_list.append(sac)
    return sacpz_list


def main(dir, drop_comment=False, m_to_nm=1, geophone=False):
    nscls = read(dir, SacPolesZeros)
    write_all(dir, nscls, drop_comment=drop_comment, m_to_nm=m_to_nm, geophone=geophone)


def main_args():
    try:
        parser = ArgumentParser(prog='ewconfig',
                                description="""A scanner program to 
take SAC PZ files from rdseed -pf and convert them to configurations suitable 
for use with localmag, hypoinverse, pick_ew, pick_FP, slink2ew, carlstatrig and
wave_serverV station lists. It is presumed that the dataless was complete and had 
lat/long/elevation for the SCNL. Output sac files will be in Meters unless the
--nano parameter is set to convert to nanometers. The output files will appear
in the current directory.""")
        parser.add_argument('pzdirectory', type=assert_dir,
                            help='Directory containing the input files -- which are the output of rdseed -pf')
        add_write_args(parser)
        add_version_args(parser)
        add_md_help_argument(parser)
        add_log_args(parser)
        args = parser.parse_args()
        make_log_from_args(parser.prog, args)
        main(args.pzdirectory, args.drop_comment, args.nano, args.geophone)
    except Exception as e:
        log.critical(e)
        exit(1)


if __name__ == '__main__':
    main_args()



from logging import getLogger
from os.path import join

from ewconfig.lib.file import assert_empty_dir

log = getLogger(__name__)

INST = 'INST_WILDCARD'
MOD = 'MOD_WILDCARD'
INDEX_SIZE = 10000
TANKSIZE_VAR = '${WSV_TANK_MEGS}'
TANKDIR_VAR = '${WSV_TANK_DIR}'
EWDATA_VAR = '${EW_DATA_DIR}'

NM_IN_M = 1e9


def write_pz_ew(dir, sncl, drop_comment=False, m_to_nm=1):
    with open(join(dir, "%s.%s.%s.%s.sac" % (sncl.net, sncl.sta, sncl.chan, sncl.loc)), "w") as f:
        if not drop_comment:
            f.write(sncl.keep_comment)
        f.write("ZEROS   %d\n" % len(sncl.zeros_lines))
        for line in sncl.zeros_lines:
            f.write(line)
        f.write("POLES   %d\n" % len(sncl.poles_lines))
        for line in sncl.poles_lines:
            f.write(line)
        sncl.constant = sncl.constant / m_to_nm
        if m_to_nm == 1:
            comment = 'UNITS IN M'
        elif m_to_nm == NM_IN_M:
            comment = 'UNITS IN NM'
        else:
            comment = f'UNITS UNKNOWN (FACTOR {m_to_nm:g})'
        f.write(f"* {comment}\nCONSTANT        {sncl.constant:e}\n")


def write_wsv(dir, sacPZs):
    with open(join(dir, 'wsv_chan_list.d'), 'w') as f:
        f.write(
            "#          names       size  (TYPE_TRACEBUF2 only)         (megabytes) (max breaks)     (full path)      Tank\n")
        for s in sacPZs:
            f.write("Tank %s %s %s %s 4096 %s %s %s %d %s/%s/%s.%s.%s.%s.wsv_tnk\n" % \
                    (s.sta, s.chan, s.net, s.loc, INST, MOD, TANKSIZE_VAR, INDEX_SIZE, EWDATA_VAR, TANKDIR_VAR, \
                     s.sta, s.chan, s.net, s.loc))


def write_pick_sta(dir, sacPZs, geophone=False):
    # example from Raton
    #    1     1  T25A  BHZ TA -- 3  40  3 162  500  3  .939  3.  .4  .015 5.  .9961  1200. 132.7   .8  1.5 135000. 8388608
    #    1     1  T25A  BHE TA -- 3  40  3 162  500  3  .939  3.  .4  .015 5.  .9961  1200. 132.7   .8  1.5 135000. 8388608
    #    1     1  T25A  BHN TA -- 3  40  3 162  500  3  .939  3.  .4  .015 5.  .9961  1200. 132.7   .8  1.5 135000. 8388608
    with open(join(dir, 'pick_sta.d'), 'w') as f:
        f.write(
            '#\n #                              MinBigZC       RawDataFilt    LtaFilt         DeadSta          PreEvent\n')
        f.write('# Pick  Pin    Station/   MinSmallZC   MaxMint           StaFilt       RmavFilt           AltCoda\n')
        f.write(
            '# Flag  Numb   Comp/Net   Itr1   MinPeakSize  i9  CharFuncFilt  EventThresh          CodaTerm         Erefs\n')
        f.write(
            '# ----  ----   --------   ----------------------------------------------------------------------------------\n')
        for s in sacPZs:
            if geophone or s.chan.startswith('E'):
                f.write("1 1 %s %s %s %s 3 40 3 300 500 0 .985 0. .0198 .002 3. .9961 1200. 100.0 .8 1.5 50000. 8388608\n" % \
                        (s.sta, s.chan, s.net, s.loc))
            else:
                f.write(
                    "1  1 %s %s %s %s 3  40  3 162  500  3  .939  3.  .4  .015 5.  .9961  1200. 132.7   .8  1.5 135000. 8388608\n" % \
                    (s.sta, s.chan, s.net, s.loc))


def write_hinv_sta(dir, sac_pzs):
    # P02   XR  EHZ  37 05.0890 104 51.1720 17830.0     0.02  0.00  0.00  0.00    0.0000
    with open(join(dir, 'hinv_sta.d'), 'w') as f:
        for s in sac_pzs:
            lat_deg = int(s.lat)
            lat_min = (s.lat - lat_deg) * 60.
            lon_deg = int(s.lon)
            lon_min = (s.lon - lon_deg) * 60.
            if lon_deg > 0:
                lon_char = "E"
            else:
                lon_char = "W"
            if lon_char == "W":
                # remove negative values
                lon_deg = lon_deg * -1.0
                lon_min = lon_min * -1.0
            f.write("%-5s %2s  %3s  %2d %7.4f %3d %7.4f %4d0.0     0.00  0.00  0.00  0.00    0.00%2s\n" % \
                    (s.sta, s.net, s.chan, lat_deg, lat_min, lon_deg, lon_min, int(s.elev), s.loc))


def write_seed_link_stas(dir, ns_dict):
    with open(join(dir, 'slink_imports.d'), 'w') as f:
        for ns in sorted(ns_dict.keys()):
            f.write("Stream   %s_%s \"%s?.D\"\n" % (ns_dict[ns].net, ns_dict[ns].sta, ns_dict[ns].chan[0:2]))


def write_pick_fp(dir, sac_pzs):
    #                                      threshold1
    # Pick  Pin     Sta/Comp           longTermWindow  tUpEvent
    # Flag  Numb    Net/Loc       filterWindow  threshold2
    # ----  ----    --------      -----------------------------
    #    1    00  AVG 2.4 6.0 10.0 8.0 0.2
    #
    # ----  ----   --------   ----------------------------------------------------------------------------------
    # 1  1 ACSO BH1 US 00  2.4 6.0 10.0 8.0 0.2
    with open(join(dir, 'pick_FP_sta.d'), 'w') as f:
        f.write('#                                      threshold1\n')
        f.write('# Pick  Pin     Sta/Comp           longTermWindow  tUpEvent\n')
        f.write('# Flag  Numb    Net/Loc       filterWindow  threshold2\n')
        f.write('# ----  ----    --------      -----------------------------\n')
        for s in sac_pzs:
            f.write("1    1 %s %s %s %s 2.4 6.0 10.0 8.0 0.2\n" % \
                    (s.sta, s.chan, s.net, s.loc))


def write_trig_sta(dir, sac_pzs):
    with open(join(dir, 'trigsta.scnl'), 'w') as f:
        for s in sac_pzs:
            f.write("TrigStation %s %s %s %s\n" % (s.sta, s.chan, s.net, s.loc))


def log_pz(nscl):
    log.info("%s lat=%f lon=%f elev=%f const=%e np=%d nz=%d" %
             (nscl, nscl.lat, nscl.lon,
              nscl.elev, nscl.constant, len(nscl.poles_lines), len(nscl.zeros_lines)))


def write_all(dir, nscls, drop_comment=False, m_to_nm=1, geophone=False):
    # we sort so that the output is constant even if different sources generate nscls in different orders
    nscls = sorted(nscls, key=str)
    chan = assert_empty_dir(join(dir, 'chan'))
    write_wsv(chan, nscls)
    write_pick_sta(chan, nscls, geophone=geophone)
    write_pick_fp(chan, nscls)
    write_hinv_sta(chan, nscls)
    write_trig_sta(chan, nscls)
    eqk_response = assert_empty_dir(join(dir, 'eqk', 'response'))
    ns_dict = {}
    for nscl in nscls:
        write_pz_ew(eqk_response, nscl, drop_comment=drop_comment, m_to_nm=m_to_nm)
        ns = nscl.net + nscl.sta
        if ns not in ns_dict:
            ns_dict[ns] = nscl
    write_seed_link_stas(chan, ns_dict)


def add_write_args(parser, with_drop_comment=True):
    if with_drop_comment:
        parser.add_argument('-d', '--drop-comment', action="store_true",
                            help='Delete the comments from the sac output.')
    else:
        parser.set_defaults(drop_comment=False)
    parser.add_argument('-n', '--nano', action='store_const',
                        const=NM_IN_M, default=1,
                        help='Convert Constant meters into nanometers. '
                             'You must use this option if you are using the default Localmag or GMEW settings. '
                             'If you have enabled the ResponseInMeters option in localmag.d and/or gmew.d, '
                             'then you don\'t need this option.')
    parser.add_argument('-g', '--geophone', action="store_true",
                        help='Treat ALL instruments as Geophones in pick_ew output. '
                             'Otherwise broadband settings are used unless channel starts with E.')




# https://chromium.googlesource.com/catapult/+/refs/heads/master/devil/devil/utils/markdown.py
# https://chromium.googlesource.com/catapult/+/refs/heads/master/LICENSE

# 2021-03-25 - acooke - removed unused code that had deprecated imports.
#                     - escaped * in help text

# Copyright 2015 The Chromium Authors. All rights reserved.

# Redistribution and use in source and binary forms, with or without
# modification, are permitted provided that the following conditions are met:
#
# * Redistributions of source code must retain the above copyright notice, this
#   list of conditions and the following disclaimer.
#
# * Redistributions in binary form must reproduce the above copyright notice,
#   this list of conditions and the following disclaimer in the documentation
#   and/or other materials provided with the distribution.
#
# * Neither the name of catapult nor the names of its
#   contributors may be used to endorse or promote products derived from
#   this software without specific prior written permission.
#
# THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS "AS IS"
# AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE
# IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
# DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT HOLDER OR CONTRIBUTORS BE LIABLE
# FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR CONSEQUENTIAL
# DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR
# SERVICES; LOSS OF USE, DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER
# CAUSED AND ON ANY THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY,
# OR TORT (INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
# OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

import argparse
import os
import re

from .args import ParagraphHelpFormatter

# A markdown code block template: https://goo.gl/9EsyRi

_CODE_BLOCK_FORMAT = '''```{language}
{code}
```
'''
_DEVIL_ROOT = os.path.abspath(
    os.path.join(os.path.dirname(__file__), '..', '..'))


def md_bold(raw_text):
    """Returns markdown-formatted bold text."""
    return '**%s**' % md_escape(raw_text, characters='*')


def md_code(raw_text, language):
    """Returns a markdown-formatted code block in the given language."""
    return _CODE_BLOCK_FORMAT.format(
        language=language or '', code=md_escape(raw_text, characters='`'))


def md_escape(raw_text, characters='*_'):
    """Escapes * and _."""

    def escape_char(m):
        return '\\%s' % m.group(0)

    pattern = '[%s]' % re.escape(characters)
    return re.sub(pattern, escape_char, raw_text)


def md_heading(raw_text, level):
    """Returns markdown-formatted heading."""
    adjusted_level = min(max(level, 0), 6)
    return '%s%s%s' % ('#' * adjusted_level, ' ' if adjusted_level > 0 else '',
                       raw_text)


def md_inline_code(raw_text):
    """Returns markdown-formatted inline code."""
    return '`%s`' % md_escape(raw_text, characters='`')


def md_italic(raw_text):
    """Returns markdown-formatted italic text."""
    return '*%s*' % md_escape(raw_text, characters='*')


def md_link(link_text, link_target):
    """returns a markdown-formatted link."""
    return '[%s](%s)' % (md_escape(link_text, characters=']'),
                         md_escape(link_target, characters=')'))


class MarkdownHelpFormatter(ParagraphHelpFormatter):
    """A really bare-bones argparse help formatter that generates valid markdown.
    This will generate something like:
    usage
    # **section heading**:
    ## **--argument-one**
    ```
    argument-one help text
    ```
    """

    # override
    def _format_usage(self, usage, actions, groups, prefix):
        usage_text = super(MarkdownHelpFormatter, self)._format_usage(
            usage, actions, groups, prefix)
        return md_code(usage_text, language=None)

    # override
    def format_help(self):
        self._root_section.heading = md_heading(self._prog, level=1)
        return md_escape(super(MarkdownHelpFormatter, self).format_help(), characters='*')

    # override
    def start_section(self, heading):
        super(MarkdownHelpFormatter, self).start_section(
            md_heading(heading, level=2))

    # override
    def _format_action(self, action):
        lines = []
        action_header = self._format_action_invocation(action)
        lines.append(md_heading(action_header, level=3))
        if action.help:
            lines.append(md_code(self._expand_help(action), language=None))
        lines.extend(['', ''])
        return '\n'.join(lines)


class MarkdownHelpAction(argparse.Action):
    def __init__(self,
                 option_strings,
                 dest=argparse.SUPPRESS,
                 default=argparse.SUPPRESS,
                 **kwargs):
        super(MarkdownHelpAction, self).__init__(
            option_strings=option_strings,
            dest=dest,
            default=default,
            nargs=0,
            **kwargs)

    def __call__(self, parser, namespace, values, option_string=None):
        parser.formatter_class = MarkdownHelpFormatter
        parser.print_help()
        parser.exit()


def add_md_help_argument(parser):
    """Adds --md-help to the given argparse.ArgumentParser.
    Running a script with --md-help will print the help text for that script
    as valid markdown.
    Args:
      parser: The ArgumentParser to which --md-help should be added.
    """
    parser.add_argument(
        '--md-help',
        action=MarkdownHelpAction,
        help='Print Markdown-formatted help text and exit.')




from argparse import ArgumentParser
from logging import getLogger
from os import rmdir
from os.path import isfile
from shutil import move, rmtree

from .lib.args import ParagraphHelpFormatter, add_version_args
from .lib.file import assert_dir, tmp_dir, assert_empty_dir
from .lib.filter import Filter
from .lib.log import add_log_args, make_log_from_args
from .lib.markdown import add_md_help_argument
from .merge.chan import merge_filtered_chan
from .merge.prefer import Prefer
from .merge.rest import copy_rest
from .merge.sac import merge_sac
from .stationxml2ew import add_stationxml2ew_args, main_single

log = getLogger(__name__)


def expand_xml(file, drop_comment=False, m_to_nm=1, geophone=False,
               lo_precision=False, default_response=None,
               sta_filter=Filter('Station'), chan_filter=Filter('Channel'), date=None):
    if isfile(file):
        log.warning(f'Treating {file} as a station.xml file and calling stationxml2ew')
        dir = tmp_dir()
        main_single(file, dir=dir, drop_comment=drop_comment, m_to_nm=m_to_nm, geophone=geophone,
                    lo_precision=lo_precision, default_response=default_response,
                    sta_filter=sta_filter, chan_filter=chan_filter, date=date)
        log.debug(f'Expanded {file} into {dir}')
        return dir, file
    else:
        return file, False


def prepare_dirs(old, news, merged=None, backup=None, force=False):
    # we always have a merged, even if --force (when it is copied to old later)
    old = assert_dir(old, 'Old directory (--old)')
    news = [(assert_dir(new, 'New directory (--new)'), delete) for new, delete in news]
    merged = merged or tmp_dir()
    merged = assert_empty_dir(merged, 'Merge directory (--merge)')
    if force:
        if backup == 'NONE':
            log.warning('No backup (--backup NONE)')
            backup = None
        else:
            backup = backup or tmp_dir()
            backup = assert_empty_dir(backup, 'Backup directory (--backup)')
            rmdir(backup)  # so that move creates this directory instead of copying to inside it
    return old, news, merged, backup


def main(old, news, merged=None, backup=None, force=False, prefer=Prefer.CHECK, sort=True,
         drop_comment=False, m_to_nm=1, geophone=False, lo_precision=False, default_response=None,
         sta_filter=Filter('Station'), chan_filter=Filter('Channel'), date=None):
    if force:
        if merged: raise Exception('Do not specify --merged with --force')
    else:
        if backup: raise Exception('Do not specify --backup without --force')
    news = [expand_xml(new, drop_comment=drop_comment, m_to_nm=m_to_nm, geophone=geophone,
                       lo_precision=lo_precision, default_response=default_response,
                       sta_filter=sta_filter, chan_filter=chan_filter, date=date) for new in news]
    old, news, merged, backup = prepare_dirs(old, news, merged=merged, backup=backup, force=force)
    for new, delete in news:
        log.info(f'Processing directory {new}')
        merge_sac(old, new, merged, prefer)
        merge_filtered_chan(old, new, merged, prefer, sort)
        if delete:
            log.debug(f'Deleting temporary {new} (where {delete} was expanded)')
            rmtree(new)
    copy_rest(old, merged)
    if force:
        if backup:
            log.info(f'Moving {old} to {backup} (creating backup)')
            move(old, backup)
        else:
            log.warning(f'Deleting {old} (--backup NONE)')
            rmtree(old)
        log.info(f'Moving {merged} to {old} (overwriting old)')
        move(merged, old)
        return backup
    else:
        return merged


def main_args():
    try:
        parser = ArgumentParser(prog='ewmerge', add_help=False,
                                formatter_class=ParagraphHelpFormatter,
                                description='''A program to merge two EW configurations 
(at --old and --new) selecting entries based on NSCLs.

In the final output, files in the eqk/response and chan sub-directories are merged / filtered 
using NSCLs as described above.  All other files and directories are taken from --old.

If --force is NOT given then the output is saved to a new directory, whose path is printed
to stdout.  This path can be specified with --merged.

If --force is given then the output replaces the --old directory.  A copy of the initial data 
in --old is made and the path to this directory printed to stdout.  This path can be specified with 
--backup.

If you really want to overwrite existing with no backup data use "--force --backup NONE".
''')
        group = parser.add_argument_group(title='Merge parameters')
        group.add_argument('--old', required=True, help='Directory for old config.')
        group.add_argument('--new', required=True, nargs='+',
                           help='Directory(s) or XML file(s) for new config.  '
                                'XML files are expanded by stationxml2ew (see parameters below).')
        prefer = group.add_mutually_exclusive_group()
        prefer.add_argument('--prefer-old', action='store_const', dest='prefer', const=Prefer.OLD,
                            help='If a NSCL is present in old and new, use the old entries '
                                 '(default is to require identical entries).')
        prefer.add_argument('--prefer-new', action='store_const', dest='prefer', const=Prefer.NEW,
                            help='If a NSCL is present in old and new, use the new entries.')
        group.set_defaults(prefer=Prefer.CHECK)
        group.add_argument('--force', action='store_true', help='Replace old directory with merged.')
        group.add_argument('--backup', help='Backup directory for old config when replacing (if --force).')
        group.add_argument('--merged', help='Destination directory for merged config (if not --force).')
        group.add_argument('--no-sort', default=False, action='store_true',
                            help='Don\'t sort NSCLs in output.')
        group.add_argument('-h', action='help', help='Show this help message and exit.')
        add_version_args(group)
        add_md_help_argument(group)
        add_log_args(group)
        group = parser.add_argument_group(title='XML expansion parameters (passed to stationxml2ew)')
        add_stationxml2ew_args(group)
        args = parser.parse_args()
        make_log_from_args(parser.prog, args)
        if (args.force and args.merged) or (not args.force and args.backup):
            raise Exception('Use --merged with --force and --backup without')
        output = main(old=args.old, news=args.new, merged=args.merged, backup=args.backup,
                      force=args.force, prefer=args.prefer, sort=(not args.no_sort),
                      drop_comment=args.drop_comment, m_to_nm=args.nano, geophone=args.geophone,
                      default_response=args.default_response, date=args.date,
                      sta_filter=Filter('Station', includes=args.include_sta, excludes=args.exclude_sta),
                      chan_filter=Filter('Channel', includes=args.include_chan, excludes=args.exclude_chan))
        if output:  # None if --backup NONE
            print(output)
    except Exception as e:
        log.critical(e)
        exit(1)


if __name__ == '__main__':
    main_args()



from enum import Enum
from logging import getLogger
from re import sub

from .pz import PolesZeros

log = getLogger(__name__)


class Response(Enum):
    DISPLACEMENT = 1
    VELOCITY = 2
    ACCELERATION = 3


class UnsupportedResponseException(Exception): pass


PZ_FMT = '\t%+.6e\t%+.6e\n'
PZ_FMT_LO = '\t%+.5e \t%+.5e \n'


class XmlPolesZeros(PolesZeros):

    def __init__(self, net, sta, chan, lo_precision=False, default_response=None):
        super().__init__()
        self.__parse(net, sta, chan, lo_precision=lo_precision, default_response=default_response)

    def __parse(self, net, sta, chan, lo_precision=False, default_response=None):
        self.lat = chan.latitude
        self.lon = chan.longitude
        self.elev = chan.elevation
        self.net = net.code
        self.sta = sta.code
        self.chan = chan.code
        self.loc = chan.location_code if chan.location_code else '--'
        # the following is based on the implementation of Response.get_sacpz which gets the same data
        # that the original code used
        # https://github.com/obspy/obspy/blob/e7b9474573e15a586cb8ce73df9bbe677f0d7421/obspy/core/inventory/response.py#L1956
        # for some reason that doesn't correct to displacement (and mike doesn't think that's a problem)
        try:
            paz = chan.response.get_paz()  # may raise an exception if there is none
            self.constant = paz.normalization_factor * chan.response.instrument_sensitivity.value
            self.zeros_lines = [self.__format_pz(c, lo_precision) for c in paz.zeros]
            self.poles_lines = [self.__format_pz(c, lo_precision) for c in paz.poles]
            response = self.__response_from_input_units(chan, default_response=default_response)
            if self.zeros_lines or self.poles_lines:
                for _ in range(response.value-1):  # shift to displacement
                    # has to be at start to match original
                    self.zeros_lines.insert(0, self.__format_pz(0j, lo_precision))
        except UnsupportedResponseException:
            raise
        except Exception as e:
            log.debug(f'Discarding {e} ({type(e)})')
            log.info(f'No PAZ data in {self}')

    def __format_pz(self, imag, lo_precision):
        # the original code printed more precision than it actually had(!)
        if lo_precision:
            text = PZ_FMT_LO % (imag.real, imag.imag)
            return sub(r'e([-+]\d{2}) ', r'0e\1', text)
        else:
            return PZ_FMT % (imag.real, imag.imag)

    def __response_from_input_units(self, chan, default_response=None):
        try:
            instrument_sensitivity = chan.response.instrument_sensitivity
            units_map = {'m': Response.DISPLACEMENT,
                         'm/s': Response.VELOCITY,
                         'm/s**2': Response.ACCELERATION}
            input_units = instrument_sensitivity.input_units.lower()
            if input_units == 'pa':
                raise UnsupportedResponseException('Pressure sensitive detector')
            if input_units in units_map:
                response = units_map[input_units]
                log.debug(f'Input units {input_units} => {response}')
                return response
            log.warning(f'Could not parse input units {input_units} for {self}')
        except UnsupportedResponseException as e:
            raise
        except Exception as e:
            log.warning(f'Could not infer response for {self} ({e})')
        if default_response:
            log.info(f'Using default response for {self}: {default_response}')
            return default_response
        else:
            raise UnsupportedResponseException(f'Cannot infer response type for {self} '
                                               f'(use --velocity etc to override))')



from logging import getLogger
from os import listdir
from os.path import join, exists
from re import compile

from .parser import BaseParser
from .prefer import Prefer
from .slink import SlinkParser

log = getLogger(__name__)

PARSERS = {}


def merge_filtered_chan(old, new, merged, prefer, sort):
    # by the time we get here we've used the filters to select NSCLs from the two directories
    all_files = set(_ls_chan(old)).union(_ls_chan(new))
    for file in sorted(all_files):
        if file in PARSERS:
            Parser = PARSERS[file]
            log.debug(f'Parsing {file} with {Parser}')
            new_parser = Parser(file, prefer, sort)
            new_parser.read(new)
            old_parser = Parser(file, prefer, sort)
            old_parser.read(old)
            old_parser.add(new_parser)
            old_parser.write(merged)
        else:
            log.error(f'No support for {file}')


def _ls_chan(dir):
    chan = join(dir, 'chan')
    if exists(chan):
        return listdir(chan)
    else:
        return []


class RegexParser(BaseParser):
    '''
    Files with an optional header where each line contains NSCL related data.
    Each line is matched against a pattern and compared with the pre-selected NSCLs.
    '''

    def __init__(self, line_regex, file, prefer, sort, header_regex=None):
        self.__line_regex = line_regex
        self.__prefer = prefer
        self.__sort = sort
        self.__header_regex = header_regex
        self.__header = []
        self.__lines = {}
        super().__init__(file)

    def _match_to_key(self, match):
        raise NotImplementedError('_match_to_key')

    def _read(self, file):
        header = True
        for line in file:
            if header:
                if self.__header_regex and self.__header_regex.match(line):
                    self.__header.append(line)
                else:
                    header = False
            if not header:
                match = self.__line_regex.match(line)
                if match:
                    key = self._match_to_key(match)
                    self.__lines[key] = line
                else:
                    raise Exception(f'Could not parse {line} in {self._file}')

    def add(self, new):
        if self.__header != new.__header:
            log.warning(f'Headers differ for {self._file}')
        for key, line in new.__lines.items():
            if key in self.__lines:
                if line != self.__lines[key]:
                    if self.__prefer == Prefer.OLD:
                        log.warning(f'Taking {key} from {self._file} (old)')
                        pass  # already have the old line
                    elif self.__prefer == Prefer.NEW:
                        log.warning(f'Taking {key} from {new._file} (new)')
                        self.__lines[key] = line
                    else:
                        raise Exception(f'Entries differ for {key} in {self._file}')
                else:
                    log.debug(f'Entries consistent for {key} in {self._file}')
            else:
                self.__lines[key] = line

    def _write(self, file):
        for line in self.__header:
            file.write(line)
        sort = sorted if self.__sort else list
        for key in sort(self.__lines.keys()):
            file.write(self.__lines[key])


class NSCL(RegexParser):
    '''
    Selection uses net, sta, chan and loc.
    '''

    def _match_to_key(self, match):
        return f'{match.group("net")}.{match.group("sta")}.{match.group("chan")}.{match.group("loc")}'


class NSC(RegexParser):
    '''
    Selection uses net, sta and chan.
    '''

    def _match_to_key(self, match):
        return f'{match.group("net")}.{match.group("sta")}.{match.group("chan")}'


class NS(RegexParser):
    '''
    Selection uses net and sta.
    '''

    def _match_to_key(self, match):
        return f'{match.group("net")}.{match.group("sta")}'


hash_header = compile(r'^\s*#')
SNC_prefix = compile(r'^\s*(?P<sta>[A-Z0-9]+)\s+(?P<net>[A-Z0-9]+)\s+(?P<chan>[A-Z0-9]+)\s+')
SCN_from_3 = compile(r'\s*\d+\s+\d+\s+(?P<sta>[A-Z0-9]+)\s+(?P<chan>[A-Z0-9]+)\s+(?P<net>[A-Z0-9]+)\s+')
NS_after_Stream =  compile(r'\s*Stream\s*(?P<net>[A-Z0-9]+)_(?P<sta>[A-Z0-9]+)\s+')
SCNL_after_TrigStation = compile(r'\s*TrigStation\s+(?P<sta>[A-Z0-9]+)\s+(?P<chan>[A-Z0-9]+)\s+(?P<net>[A-Z0-9]+)\s+(?P<loc>(?:[A-Z0-9]+|--))')
SCNL_after_Tank = compile(r'\s*Tank\s+(?P<sta>[A-Z0-9]+)\s+(?P<chan>[A-Z0-9]+)\s+(?P<net>[A-Z0-9]+)\s+(?P<loc>(?:[A-Z0-9]+|--))')


class HinvSta(NSC):

    def __init__(self, file, prefer, sort):
        super().__init__(SNC_prefix, file, prefer, sort)


class PickSta(NSC):

    def __init__(self, file, prefer, sort):
        super().__init__(SCN_from_3, file, prefer, sort, hash_header)


class SlinkImports(NS):
    '''
    Old code, no longer used.  File contents are themselves patterns and must be applied to the NSCLS.
    See new code in slink.py
    '''

    def __init__(self, file, prefer, sort):
        super().__init__(NS_after_Stream, file, prefer, sort)


class TrigSta(NSCL):

    def __init__(self, file, prefer, sort):
        super().__init__(SCNL_after_TrigStation, file, prefer, sort)


class WsvChanList(NSCL):

    def __init__(self, file, prefer, sort):
        super().__init__(SCNL_after_Tank, file, prefer, sort, hash_header)


PARSERS['hinv_sta.d'] = HinvSta
PARSERS['pick_FP_sta.d'] = PickSta
PARSERS['pick_sta.d'] = PickSta
PARSERS['slink_imports.d'] = SlinkParser
PARSERS['trigsta.scnl'] = TrigSta
PARSERS['wsv_chan_list.d'] = WsvChanList



import datetime as dt

ALL_DATE_FORMATS = ('%Y-%m-%dT%H:%M:%S.%f', '%Y-%m-%d %H:%M:%S.%f',
                    '%Y-%m-%dT%H:%M:%S', '%Y-%m-%d %H:%M:%S',
                    '%Y-%m-%dT%H:%M', '%Y-%m-%d %H:%M',
                    '%Y-%m-%d')


def parse_date(value):
    for format in ALL_DATE_FORMATS:
        try:
            return dt.datetime.strptime(value, format).replace(tzinfo=dt.timezone.utc)
        except ValueError:
            pass
    raise ValueError('Cannot parse "%s" as a datetime' % value)


def between(start, date, finish):
    date = date if date else dt.datetime.now()
    return (start is None or start <= date) and (finish is None or date <= finish)




from shutil import rmtree
from unittest import TestCase

from ewconfig.lib.filter import Filter
from ewconfig.pz.xml import Response
from ..lib.x2ew import MainMixin
from ..lib.log import LogMixin


class TestBadUnits(MainMixin, LogMixin, TestCase):

    stderr_verbosity = 5

    def test_station_xml(self):
        dest_dir = self.run_xml2ew('BAD-UNITS', chan_filter=Filter('Channel', excludes=['LOG']),
                                   lo_precision=True)
        with self.assertRaises(AssertionError):
            self.assert_files(dest_dir, 'BAD-UNITS',
                              skip_files=('CI.PASC.VDO.00.sac', 'CI.PASC.VKO.00.sac'),
                              skip_lines=(r'PASC .*VDO', r'PASC .*VKO'))

    def test_station_xml_with_default(self):
        dest_dir = self.run_xml2ew('BAD-UNITS',
                                   chan_filter=Filter('Channel', excludes=['LOG']), lo_precision=True,
                                   default_response=Response.ACCELERATION)
        self.assert_files(dest_dir, 'BAD-UNITS',
                          skip_files=('CI.PASC.VDO.00.sac', 'CI.PASC.VKO.00.sac'),
                          skip_lines=(r'PASC .*VDO', r'PASC .*VKO'))
        rmtree(dest_dir)



from logging import getLogger
from shutil import copytree

log = getLogger(__name__)


def copy_rest(old, merged):
    log.info(f'Copying rest of {old} to {merged}')
    copytree(old, merged, ignore=_copy_filter(old), dirs_exist_ok=True)


def _copy_filter(old):

    def filter(dir, files):
        if dir == old and 'chan' in files:
            return ['chan']
        elif dir == 'eqk' and 'response' in files:
            return ['response']
        else:
            return []

    return filter



from itertools import zip_longest
from logging import getLogger
from os import listdir
from os.path import join, exists
from re import search, sub, compile

log = getLogger(__name__)


class Compare:

    def _test_dir(self, name, subdir=None):
        raise NotImplementedError('_test_dir')

    def _assert_files(self, dest_dir, test_dir, rdseed=False, skip_files=None, skip_lines=None, extra_dirs=tuple()):
        skip_files_all = tuple() if rdseed else (r'rdseed\.err_log', r'SAC_PZs_.*')
        if skip_files: skip_files_all = skip_files_all + skip_files
        skip_lines_all = (r'^Date', r'^CREATED', r'^\*')
        if skip_lines: skip_lines_all = skip_lines_all + skip_lines
        for subdir in ('chan', join('eqk', 'response'), *extra_dirs):
            self.__assert_dirs(join(test_dir, subdir), join(dest_dir, subdir), skip_files=skip_files_all)
            self.__assert_dirs(join(dest_dir, subdir), join(test_dir, subdir), contents=True,
                               skip_files=skip_files_all, skip_lines=skip_lines_all)

    def __assert_dirs(self, dir1, dir2, contents=False, skip_files=tuple(), skip_lines=tuple()):
        for file in listdir(dir1):
            if any(search(skip, file) for skip in skip_files):
                log.debug(f'Skipping {file}')
            else:
                path1 = join(dir1, file)
                path2 = join(dir2, file)
                self.assertTrue(exists(path2), f'{file} from {dir1} missing from {dir2}')
                if contents:
                    with open(path1) as file1, open(path2) as file2:
                        lines1 = [line for line in file1 if not any(search(skip, line) for skip in skip_lines)]
                        lines2 = [line for line in file2 if not any(search(skip, line) for skip in skip_lines)]
                        for line1, line2 in zip_longest(lines1, lines2):
                            line1 = self.__fix_negative_zero(line1)
                            line2 = self.__fix_negative_zero(line2)
                            if line1 and line2 and 'CONSTANT' in line1 and 'CONSTANT' in line2:
                                self.__compare_constants(line1, line2)
                            else:
                                self.assertEqual(line1, line2, f'from {file} ({dir1} v {dir2})')

    def __fix_negative_zero(self, line):
        # we don't care about -0 being different to 0
        if line:
            return sub(r'\-0\.000000e\+00', r'0.000000e+00', line)
        else:
            return line

    def __compare_constants(self, line1, line2):
        # need to drop some precision here.  suspect it is ok.
        CONSTANT = compile(r'CONSTANT\s+(\S*)\s*(?:\*.*)?')  # optional group is trailing comment (units)
        constant1 = '%.4e' % float(CONSTANT.match(line1).group(1))
        constant2 = '%.4e' % float(CONSTANT.match(line2).group(1))
        self.assertEqual(constant1, constant2, f'{line1} / {line2}')



from os.path import join
from shutil import rmtree
from unittest import TestCase

from ewconfig.lib.filter import Filter
from ewconfig.lib.write import NM_IN_M
from ..lib.x2ew import MainMixin
from ..lib.log import LogMixin


class TestCiPasc(MainMixin, LogMixin, TestCase):

    stderr_verbosity = 5

    def test_original(self):
        dest_dir = self.run_sac2ew('CI-PASC')
        self.assert_files(dest_dir, 'CI-PASC', rdseed=True)
        rmtree(dest_dir)

    def test_station_xml(self):
        dest_dir = self.run_xml2ew('CI-PASC', chan_filter=Filter('Channel', excludes=['LOG']), lo_precision=True)
        # ignore pressure sensitive channels which are now excluded
        self.assert_files(dest_dir, 'CI-PASC',
                          skip_files=('CI.PASC.VDO.00.sac', 'CI.PASC.VKO.00.sac'),
                          skip_lines=(r'PASC .*VDO', r'PASC .*VKO'))
        rmtree(dest_dir)

    def test_station_xml_nm(self):
        dest_dir = self.run_xml2ew('CI-PASC', chan_filter=Filter('Channel', excludes=['LOG']), lo_precision=True,
                                   m_to_nm=NM_IN_M)
        # everything else has already been tested in previous test
        with open(join(dest_dir, 'eqk/response/CI.PASC.BHE.00.sac')) as sac:
            text = sac.read()
            self.assertTrue('* UNITS IN NM\nCONSTANT        2.925447e+04' in text, text)
        rmtree(dest_dir)



from glob import glob
from logging import getLogger
from os.path import join, basename, sep
from re import compile, escape
from shutil import copy

from .merge import merge
from ..lib.file import assert_dir, assert_equal_contents

log = getLogger(__name__)


def merge_sac(old, new, merged, prefer):
    merge(old, new, merged, prefer, parse_sac, assert_equal_contents, write_sac)


def parse_sac(dir):
    return {nscl: path for nscl, path in glob_sac(dir)}


def glob_sac(dir):
    if dir:
        eqk_response = assert_dir(join(dir, 'eqk', 'response'))
        for file in glob(join(eqk_response, '*.sac')):
            match = NSCL_sac.match(file)
            if match:
                yield match.group(1), file
            else:
                log.warning(f'Could not parse {file}')


def write_sac(merged, nscls_map):
    dest_eqk_response = assert_dir(join(merged, 'eqk', 'response'), create=True)
    for nscl, path in nscls_map.items():
        log.debug(f'Copying {path} for {nscl} to {dest_eqk_response}')
        copy(path, join(dest_eqk_response, basename(path)))


NSCL_sac = compile(r'^.*' + escape(sep) + r'([A-Z0-9]+\.[A-Z0-9]+\.[A-Z0-9]+\.[-A-Z0-9]+)\.sac$')


from argparse import SUPPRESS, HelpFormatter
from re import split


class ParagraphHelpFormatter(HelpFormatter):
    '''
    Preserve paragraphs in the description and epilog.
    '''

    def add_text(self, text):
        if text is not SUPPRESS and text is not None:
            for para in split(r'\n\s*\n', text):
                self._add_item(self._format_text, [para])


def add_version_args(parser):
    from .. import __version__
    parser.add_argument('-V', '--version', action='version', version=__version__,
                        help='Show program\'s version number and exit.')




from logging import getLogger, DEBUG, Formatter, StreamHandler
from logging.handlers import RotatingFileHandler
from os.path import join, exists

# need to save these to close logging
# (so we can restart at different levels for different tests)
__FILE_HANDLER = None
__STDERR_HANDLER = None


def make_log(progname, verbosity, log_verbosity=4, path=None):

    global __FILE_HANDLER, __STDERR_HANDLER

    level = 10 * (6 - verbosity)
    log_level = 10 * (6 - log_verbosity)

    if not getLogger(progname).handlers:

        sacpz2ew_log = getLogger('ewconfig')
        sacpz2ew_log.setLevel(DEBUG)
        test_log = getLogger('tests')
        test_log.setLevel(DEBUG)
        # capture logging from an executing module, if one exists
        main_log = getLogger('__main__')
        main_log.setLevel(DEBUG)

        if path:
            filepath = join(path, progname + '.log')
            old_log = exists(filepath)
            file_formatter = Formatter('%(levelname)-8s %(asctime)s: %(message)s')
            try:
                __FILE_HANDLER = RotatingFileHandler(filepath, backupCount=10)
            except Exception as e:
                raise Exception(f'\nThere was an error trying to create logs at {filepath} '
                                f'and so logging failed to initialise.'
                                f'\n\n({e})\n')
            # we rollover so that the new log goes to a new file
            # since we specify no size limit the entire load goes into a single log
            if old_log: __FILE_HANDLER.doRollover()
            __FILE_HANDLER.setLevel(log_level)
            __FILE_HANDLER.setFormatter(file_formatter)
            sacpz2ew_log.addHandler(__FILE_HANDLER)
            test_log.addHandler(__FILE_HANDLER)
            main_log.addHandler(__FILE_HANDLER)

        if verbosity:
            stderr_formatter = Formatter('%(levelname)8s: %(message)s')
            __STDERR_HANDLER = StreamHandler()
            __STDERR_HANDLER.setLevel(level)
            __STDERR_HANDLER.setFormatter(stderr_formatter)
            test_log.addHandler(__STDERR_HANDLER)
            sacpz2ew_log.addHandler(__STDERR_HANDLER)
            main_log.addHandler(__STDERR_HANDLER)


def clear_log():
    global __FILE_HANDLER, __STDERR_HANDLER
    if __FILE_HANDLER or __STDERR_HANDLER:
        for handler in __FILE_HANDLER, __STDERR_HANDLER:
            for log in 'ewconfig', 'tests', '__main__':
                getLogger(log).removeHandler(handler)
        __FILE_HANDLER.close()
        __FILE_HANDLER = None
        __STDERR_HANDLER = None


def add_log_args(parser):
    parser.add_argument('-v', '--verbosity', default=4, type=int, metavar='INTEGER',
                        help='Log level (0: silent, 5: debug)')


def make_log_from_args(prog, args):
    make_log(prog, args.verbosity)



from abc import ABC
from logging import getLogger
from os import makedirs
from os.path import join, exists

log = getLogger(__name__)


class BaseParser(ABC):

    def __init__(self, file):
        self._file = file

    def read(self, dir):
        path = join(dir, 'chan', self._file)
        if exists(path):
            log.debug(f'Reading {path}')
            with open(path, 'r') as src:
                self._read(src)
        else:
            log.warning(f'Could not find {path}')

    def _read(self, file):
        raise NotImplementedError('_read')

    def add(self, new):
        raise NotImplementedError('add')

    def write(self, merged):
        dir = join(merged, 'chan')
        if not exists(dir):
            makedirs(dir)
        path = join(dir, self._file)
        log.info(f'Writing {path}')
        with open(path, 'w') as file:
            self._write(file)

    def _write(self, file):
        raise NotImplementedError('_write')




from itertools import zip_longest
from logging import getLogger
from os import listdir, rmdir
from os.path import join, exists
from re import search
from shutil import copytree

from ewconfig.lib.file import tmp_dir
from .compare import Compare
from ..lib.file import project_root

log = getLogger(__name__)


class MainMixin(Compare):

    def _src_dir(self, name):
        # use the conversion test targets as input
        return join(project_root(), 'data', name.lower(), 'x2ew')

    def _src_dir_or_xml(self, name):
        if name.endswith('.xml'):
            return join(project_root(), 'data', name.split('.')[0].lower(), name)
        else:
            return self._src_dir(name)

    def _test_dir(self, name, subdir):
        return join(project_root(), 'data', name.lower(), subdir)

    def run_merge(self, old, news, force=False, **kwargs):
        from ewconfig.ewmerge import main
        old_dir = self._src_dir(old)
        new_dirs = [self._src_dir_or_xml(new) for new in news]
        if force:
            # need to copy old_dir since we don't want to actually overwrite it
            tmp_old_dir = tmp_dir()
            rmdir(tmp_old_dir)
            copytree(old_dir, tmp_old_dir)
            # returns backup, old_dir and tmp_old_dir (merged, which needs to be passed to assert_files)
            # (so old_dir is the original, tmp_old_dir is a copy we made so that overwriting wouldn't
            # erase the test data, and backup is the backed-up copy of tmp_old_dir)
            return main(tmp_old_dir, new_dirs, force=True, **kwargs), old_dir, tmp_old_dir
        else:
            # returns merged
            return main(old_dir, new_dirs, force=False, **kwargs)

    def assert_files(self, merged, name, subdir='merge', skip_files=None, skip_lines=None,
                     extra_dirs=tuple()):
        test_dir = self._test_dir(name, subdir)
        self._assert_files(merged, test_dir, skip_files=skip_files, skip_lines=skip_lines, extra_dirs=extra_dirs)



from argparse import ArgumentParser
from logging import getLogger
from os import getcwd

from obspy import read_inventory
from obspy.io.stationxml.core import validate_stationxml

from .lib.args import ParagraphHelpFormatter, add_version_args
from .lib.date import parse_date, between
from .lib.file import assert_dir
from .lib.filter import Filter
from .lib.log import add_log_args, make_log_from_args
from .lib.markdown import add_md_help_argument
from .lib.write import write_all, add_write_args
from .pz.xml import XmlPolesZeros, Response, UnsupportedResponseException

'''
The equivalent of sacpz2ew, but takes input from station.xml files. 
'''

log = getLogger(__name__)


def _validate(file):
    ok, errors = validate_stationxml(file)
    if not ok:
        log.info('The following warnings are from XML validation:')
        for error in errors:
            log.warning(error)
        # paul's Ince_v1 file triggers warnings
        log.info('End of XML validation warnings')
        log.info('Will continue despite warnings above')


def _parse(metadata, sta_filter, chan_filter, date=None, lo_precision=False, default_response=None):
    for network in metadata.networks:
        for station in sta_filter(network.stations, key=lambda sta: sta.code):
            for channel in chan_filter(station.channels, key=lambda chan: chan.code):
                nscl = f'{network.code}.{station.code}.{channel.code}.' \
                       f'{channel.location_code if channel.location_code else "--"}'
                if between(channel.start_date, date, channel.end_date):
                    try:
                        log.info(f'Processing {nscl}')
                        yield XmlPolesZeros(network, station, channel,
                                            lo_precision=lo_precision, default_response=default_response)
                    except UnsupportedResponseException as e:
                        log.warning(f'Skipping {nscl}: {e}')
                else:
                    log.warning(f'Skipping {nscl}: date range {channel.start_date} - {channel.end_date}')


def main_multiple(files, dirs=None, drop_comment=False, m_to_nm=1, geophone=False,
                  lo_precision=False, default_response=None,
                  sta_filter=Filter('Station'), chan_filter=Filter('Channel'), date=None):
    if len(files) == 1:
        if len(dirs) == 0:
            dirs = [None]
    if len(dirs) != len(files):
        raise Exception('The number of output directories must match the number of XML files.')
    for file, dir in zip(files, dirs):
        log.info(f'Expanding {file} into {dir}')
        main_single(file, dir=dir, drop_comment=drop_comment, m_to_nm=m_to_nm, geophone=geophone,
                    lo_precision=lo_precision, default_response=default_response,
                    sta_filter=sta_filter, chan_filter=chan_filter, date=date)


def main_single(file, dir=None, drop_comment=False, m_to_nm=1, geophone=False,
                lo_precision=False, default_response=None,
                sta_filter=Filter('Station'), chan_filter=Filter('Channel'), date=None):
    log.debug(f'Calling stationxml2ew single for file={file} dir={dir} drop_comment={drop_comment} '
              f'm_to_nm={m_to_nm} geophone={geophone} lo_precision={lo_precision} '
              f'default_response={default_response}sta_filter={sta_filter} chan_filter={chan_filter} '
              f'date={date}')
    _validate(file)
    dir = assert_dir(dir or getcwd(), 'Output dir (--dir)', create=True)
    metadata = read_inventory(file)
    scnls = list(_parse(metadata, sta_filter, chan_filter, date=date,
                        lo_precision=lo_precision, default_response=default_response))
    write_all(dir, scnls, drop_comment=drop_comment, m_to_nm=m_to_nm, geophone=geophone)
    
    
def add_stationxml2ew_args(parser):
    # separated here so that it can also be used by ewmerge
    parser.add_argument('--include-chan', metavar='PATTERN', nargs='*',
                        help='Pattern to match included channel names (default all).')
    parser.add_argument('--exclude-chan', metavar='PATTERN', nargs='*',
                        help='Pattern to match excluded channel names.')
    parser.add_argument('--include-sta', metavar='PATTERN', nargs='*',
                        help='Pattern to match included station names (default all).')
    parser.add_argument('--exclude-sta', metavar='PATTERN', nargs='*',
                        help='Pattern to match excluded station names.')
    parser.add_argument('--date', type=parse_date, help='Use channels valid on this date (default now).')
    defaults = parser.add_mutually_exclusive_group()
    defaults.add_argument('--displacement', dest='default_response', action='store_const',
                          const=Response.DISPLACEMENT, help='Assume unrecognized responses are displacement.')
    defaults.add_argument('--velocity', dest='default_response', action='store_const',
                          const=Response.VELOCITY, help='Assume unrecognized responses are velocity.')
    defaults.add_argument('--acceleration', dest='default_response', action='store_const',
                          const=Response.ACCELERATION, help='Assume unrecognized responses are acceleration.')
    parser.set_defaults(default_response=None)
    add_write_args(parser, with_drop_comment=False)


def main_args():
    try:
        parser = ArgumentParser(prog='stationxml2ew',
                                formatter_class=ParagraphHelpFormatter,
                                description='''A scanner program to 
convert a station.xml file to configurations suitable for use with localmag, hypoinverse, 
pick_ew, pick_FP, slink2ew, carlstatrig and wave_serverV station lists. It is presumed 
that the input file is complete, with lat/long/elevation for the SCNL.  Output sac files 
will be in meters unless the --nano parameter is set to convert to nanometers. 

Stations and channels can be selected using --include-chan (-sta) and --exclude-chan (-sta).
These take patterns that can match any string (*), any character (?), and any character from 
a sequence ([abc]).  If --include-chan (-sta) is not given it is assumed to be "*".  Channels,
for example, are included if they match --include-chan and do not match --exclude-chan.

Specifying --date selects channels active at that time.  By default this is the current time.

Responses types are recognised via the response units.  Unrecognised units can be given a default
using --displacement|--velocity|--acceleration.  If no default is given they are skipped.  Pressure
sensitive stations (units Pa) are always skipped.    
''')
        parser.add_argument('file', nargs='+', help='Station XML file(s)')
        parser.add_argument('--dir', nargs='*', default=[], metavar='DIR',
                            help='Output directory(s) (default CWD for a single input file)')
        add_stationxml2ew_args(parser)
        add_version_args(parser)
        add_md_help_argument(parser)
        add_log_args(parser)
        args = parser.parse_args()
        make_log_from_args(parser.prog, args)
        main_multiple(args.file, dirs=args.dir,
                      drop_comment=args.drop_comment, m_to_nm=args.nano, geophone=args.geophone,
                      default_response=args.default_response, date=args.date,
                      sta_filter=Filter('Station', includes=args.include_sta, excludes=args.exclude_sta),
                      chan_filter=Filter('Channel', includes=args.include_chan, excludes=args.exclude_chan))
    except Exception as e:
        # log.exception(e)
        log.critical(e)
        exit(1)


if __name__ == '__main__':
    main_args()



from logging import getLogger

from .parser import BaseParser
from .prefer import Prefer

'''
https://ds.iris.edu/ds/nodes/dmc/manuals/slink2ew/

#
Selectors      "BH?.D"           # SeedLink selectors.  These selectors are used
                                 # for a uni-station mode connection.  If one
                                 # or more 'Stream' entries are given these are
                                 # used as default selectors for multi-station
                                 # mode data streams.  See description of
                                 # SeedLink selectors below.  Multiple selectors
                                 # must be enclosed in quotes.
#
#
# List each data stream (a network and station code pair) that you
# wish to request from the server with a "Stream" command.  If one or
# more Stream commands are given the connection will be configured in
# multi-station mode (multiple station data streams over a single
# network connection).  If no Stream commands are specified the
# connection will be configured in uni-station mode, optionally using
# any specified "Selectors".  A Stream command should be followed by a
# stream key, a network code followed by a station code separated by
# an underscore (i.e. IU_KONO).  SeedLink selectors for a specific
# stream may optionally be specified after the stream key.  Multiple
# selectors must be enclosed in quotes.  Any selectors specified with
# the Selectors command above are used as defaults when no selectors
# are specified for a given stream.
#
# Combined with the above specified default selectors the BHx channels
# will be requested for each of these stations except for DSB for which
# we also want the LHx channels.
#
Stream  GE_DSB   "BH?.D LH?.D"
Stream  GE_ISP
Stream  GE_APE
Stream  GE_STU
#
#
# Some SeedLink servers support extended selection capability allowing
# wildcards (either '*' or '?') in the network and station fields, for example
# to select all stations from the TA network:
#
Stream TA_*
'''

log = getLogger(__name__)


class SlinkParser(BaseParser):
    '''
    To keep life simple:
    * Comments are not propagated if they are on the same line as data that can be merged.
    * Selectors and Streams are moved to the bottom of the file.
    * Tests for conflicts between old and new ignore patterns (treat everything as text)
    '''

    def __init__(self, file, prefer, sort):
        super().__init__(file)
        self.__prefer = prefer
        self.__sort = sort
        self.__other = []
        # not private only to ease testing
        self._selectors = []
        self._streams = {}

    def _read(self, file):
        for line in file:
            if line.strip().startswith('Stream'):
                self.__read_stream(line)
            elif line.strip().startswith('Selectors'):
                self.__read_selectors(line)
            else:
                self.__other.append(line)

    def __read_stream(self, line):
        tokens = self.__parse_tokens(line)
        if len(tokens) < 2:
            raise Exception(f'Cannot parse "{line}" in {self._file}')
        tokens = tokens[1:]
        if tokens:
            if tokens[0] in self._streams:
                raise Exception(f'Duplicate Stream {tokens[0]} in {self._file}')
            self._streams[tokens[0]] = tokens[1:]

    def __read_selectors(self, line):
        # since this is the global default there should presumably only be one line
        # but rather than throw an error on multiple lines we'll simply append to a single list
        tokens = self.__parse_tokens(line)
        if len(tokens) < 2:
            raise Exception(f'Cannot parse "{line}" in {self._file}')
        tokens = tokens[1:]
        if tokens:
            log.debug(f'Default selectors {", ".join(tokens)} from {self._file}')
            self._selectors.extend(tokens)
        else:
            log.warning(f'Lost all default selectors from {self._file}')

    def __parse_tokens(self, line):
        # all we do here is strip quotes, break into words, and quit if there's a comment
        tokens = []
        token = ''
        for char in line:
            if char in (' ', '\t', '\n', '"', "'"):
                if token:
                    tokens.append(token)
                token = ''
            elif char == '#':
                break
            else:
                token += char
        if token:
            tokens.append(token)
        return tokens

    def add(self, new):
        # the only conflicts we resolve (check) are in streams
        prev_selectors = " ".join(self._selectors)
        for selector in new._selectors:
            if selector not in self._selectors:  # avoid sets to keep ordering
                self._selectors.extend(new._selectors)
        next_selectors = " ".join(self._selectors)
        # warn about this because it applies to all streams
        # an alternative would be to move the default to the streams before merge
        if prev_selectors != next_selectors:
            log.warning(f'Default Selectors for {self._file} changed from '
                        f'{prev_selectors} to {next_selectors}')
        for net_sta, selectors in new._streams.items():
            if net_sta in self._streams:
                if self.__prefer == Prefer.OLD:
                    log.warning(f'Taking {net_sta} from {self._file} (old)')
                    pass  # already have the old data
                elif self.__prefer == Prefer.NEW:
                    log.warning(f'Taking {net_sta} from {new._file} (new)')
                    self._streams[net_sta] = selectors
                elif set(selectors) != set(self._streams[net_sta]):
                    raise Exception(f'Inconsistent selectors for Stream {net_sta} in {self._file}')
            else:
                self._streams[net_sta] = selectors

    def _write(self, file):
        for line in self.__other:
            file.write(line)
        if self._selectors:
            file.write(f'Selectors    "{" ".join(self._selectors)}"\n')
        sort = sorted if self.__sort else list
        for net_sta in sort(self._streams.keys()):
            selectors = self._streams[net_sta]
            if selectors:
                file.write(f'Stream   {net_sta} "{" ".join(selectors)}"\n')
            else:
                file.write(f'Stream   {net_sta}\n')




class PolesZeros:

    def __init__(self):
        self.lat = 0.0
        self.lon = 0.0
        self.elev = 0.0
        self.net = None
        self.sta = None
        self.chan = None
        self.loc = None
        self.constant = 0.0
        self.zeros_lines = []
        self.poles_lines = []
        self.keep_comment = ''

    def __str__(self):
        return f'{self.net}.{self.sta}.{self.chan}.{self.loc}'



import setuptools

setuptools.setup(name='ewconfig',
                 packages=setuptools.find_packages(),
                 version='0.1.4',
                 author='andrew cooke',
                 author_email='andrewcooke@isti.com',
                 description='Create EW config from Station XML',
                 long_description='''
# ewconfig

Create EW config from Station XML
                 ''',
                 long_description_content_type='text/markdown',
                 include_package_data=True,
                 setup_requires=[
                     # this is required by obspy for install
                     'numpy'
                 ],
                 install_requires=[
                     'obspy'
                 ],
                 entry_points={
                     'console_scripts': [
                         'sacpz2ew = ewconfig.sacpz2ew:main_args',
                         'stationxml2ew = ewconfig.stationxml2ew:main_args',
                         'ewmerge = ewconfig.ewmerge:main_args',
                     ],
                 },
                 classifiers=(
                     "Programming Language :: Python :: 3.9",
                     "Operating System :: OS Independent",
                     "Development Status :: 4 - Beta",
                 ),
                 )



from filecmp import dircmp
from logging import getLogger
from re import search
from shutil import rmtree
from unittest import TestCase

from ewconfig.lib.filter import Filter
from ..lib.log import LogMixin, CaptureLog
from ..lib.merge import MainMixin

log = getLogger(__name__)


class TestR50A(MainMixin, LogMixin, TestCase):

    stderr_verbosity = 5

    def test_merge_ci_pasc(self):
        merged = self.run_merge('R50A', ['CI-PASC'])
        self.assert_files(merged, 'R50A', subdir='merge_ci_pasc', extra_dirs=('extra_dir',))
        rmtree(merged)

    def test_merge_ci_pasc_xml(self):
        # tweak various things because station xml and sacp conversion are not perfectly identical
        merged = self.run_merge('R50A', ['CI-PASC.xml'], chan_filter=Filter('Channel', excludes=['LOG']),
                                lo_precision=True)
        # ignore pressure sensitive channels which are now excluded
        self.assert_files(merged, 'R50A', subdir='merge_ci_pasc', extra_dirs=('extra_dir',),
                          skip_files=('CI.PASC.VDO.00.sac', 'CI.PASC.VKO.00.sac'),
                          skip_lines=(r'PASC .*VDO', r'PASC .*VKO'))
        rmtree(merged)

    def test_merge_duplicates(self):
        with CaptureLog() as log:
            merged = self.run_merge('R50A', ['R50A'], lo_precision=True)
        # we should ONLY check the ambiguous channel
        self.assertTrue(search(r'Compared .*/N4\.R50A\.HH1\.00.sac', log.out), log.out)
        self.assertTrue(search(r'Compared .*/N4\.R50A\.HH2\.00.sac', log.out), log.out)
        self.assertTrue(search(r'Compared .*/N4\.R50A\.HHZ\.00.sac', log.out), log.out)
        self.assert_files(merged, 'R50A', subdir='merge_duplicates', extra_dirs=('extra_dir',))
        rmtree(merged)

    def test_merge_ci_pasc_forced(self):
        backup, old, merged = self.run_merge('R50A', ['CI-PASC'], force=True)
        self.assert_files(merged, 'R50A', subdir='merge_ci_pasc', extra_dirs=('extra_dir',))
        rmtree(merged)
        log.info(f'{old} v {backup}')
        cmp = dircmp(old, backup)
        cmp.report_full_closure()
        self.assertFalse(cmp.left_only)
        self.assertFalse(cmp.right_only)
        self.assertFalse(cmp.diff_files)
        self.assertFalse(cmp.funny_files)
        rmtree(backup)

    def test_merge_ci_pasc_forced_no_backup(self):
        backup, old, merged = self.run_merge('R50A', ['CI-PASC'], force=True, backup='NONE')
        self.assert_files(merged, 'R50A', subdir='merge_ci_pasc', extra_dirs=('extra_dir',))
        rmtree(merged)
        self.assertEqual(None, backup)




from logging import getLogger
from os.path import dirname, abspath, join

log = getLogger(__name__)


def project_root():
    own_dir = dirname(__file__)
    log.debug(f'lib located in {own_dir}')
    return abspath(join(own_dir, '..', '..', '..'))


